#include "mv.hh"

#include <iostream>
#include <stdlib.h> // for rand() function

using namespace std ;

// initialize structure to a matrix of size 0 x 0
void
Initialize( struct Mat & mat ) {
  mat.nrow         = 0 ;
  mat.ncol         = 0 ;
  mat.memory_block = nullptr ;
  mat.values       = nullptr ;
}

// alocate a matrix to the size: nrow x ncol
void
Allocate( struct Mat & mat, int nrow, int ncol  ) {

  if ( mat.nrow == nrow &&  mat.ncol == ncol ) return ; // nothing to do!

  mat.nrow = nrow ;
  mat.ncol = ncol ;

  if ( mat.memory_block != nullptr ) delete [] mat.memory_block ;
  mat.memory_block = new double [nrow*ncol] ;

  if ( mat.values != nullptr ) delete [] mat.values ;
  mat.values = new double* [nrow] ;

  // initialize the pointer of pointer
  for ( int i = 0 ; i < nrow ; ++i )
    mat.values[i] = mat.memory_block + i * ncol ;
}

void
AllocateAndInitializeRandom( struct Mat & mat, int nrow, int ncol ) {
  Allocate( mat, nrow, ncol ) ;
  for ( int i = 0 ; i < mat.nrow ; ++i )
    for ( int j = 0 ; j < mat.ncol ; ++j )
      mat.values[i][j] = rand() - rand() ;
}

void
Fill( struct Mat & mat, double val ) {
  for ( int i = 0 ; i < mat.nrow ; ++i )
    for ( int j = 0 ; j < mat.ncol ; ++j )
      mat.values[i][j] = val ;
}

void
Mat_x_Vec( struct Mat const & mat, // ensure the compiler that mat is not modified inside...
           struct Vec const & vec,
           struct Vec       & res ) {
  // check matrix vector compatibility
  if ( mat.ncol != vec.size ) {
    cerr << "Mat_x_Vec, incompatible dimensions\n" ;
    exit(0) ;
  }   
  if ( res.size != mat.nrow ) Allocate( res, mat.nrow ) ;
  for ( int i = 0 ; i < mat.nrow ; ++i ) {
    res.values[i] = 0 ;
    for ( int j = 0 ; j < mat.ncol ; ++i )
      res.values[i] += mat.values[i][j]*vec.values[j] ;
  }
}

void
Mat_x_Mat_123( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_123, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  Allocate( C, A.nrow, B.ncol ) ;

  // set to 0
  Fill( C, 0 ) ; // 0 int --> convertex 0 double
  
  for ( int i = 0 ; i < A.nrow ; ++i ) // 1
    for ( int j = 0 ; j < B.ncol ; ++j ) // 2
      for ( int k = 0 ; k < A.ncol ; ++k ) // 3
        C.values[i][j] += A.values[i][k]*B.values[k][j] ;
}

void
Mat_x_Mat_132( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_132, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  Allocate( C, A.nrow, B.ncol ) ;

  // set to 0
  Fill( C, 0 ) ; // 0 int --> convertex 0 double
  
  for ( int i = 0 ; i < A.nrow ; ++i ) // 1
    for ( int k = 0 ; k < A.ncol ; ++k ) // 3
      for ( int j = 0 ; j < B.ncol ; ++j ) // 2
        C.values[i][j] += A.values[i][k]*B.values[k][j] ;
}

void
Mat_x_Mat_213( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_213, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  Allocate( C, A.nrow, B.ncol ) ;

  // set to 0
  Fill( C, 0 ) ; // 0 int --> convertex 0 double
  
  for ( int j = 0 ; j < B.ncol ; ++j ) // 2
    for ( int i = 0 ; i < A.nrow ; ++i ) // 1
      for ( int k = 0 ; k < A.ncol ; ++k ) // 3
        C.values[i][j] += A.values[i][k]*B.values[k][j] ;
}

void
Mat_x_Mat_231( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_231, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  Allocate( C, A.nrow, B.ncol ) ;

  // set to 0
  Fill( C, 0 ) ; // 0 int --> convertex 0 double
  
  for ( int j = 0 ; j < B.ncol ; ++j ) // 2
    for ( int k = 0 ; k < A.ncol ; ++k ) // 3
      for ( int i = 0 ; i < A.nrow ; ++i ) // 1
        C.values[i][j] += A.values[i][k]*B.values[k][j] ;
}

void
Mat_x_Mat_312( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_312, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  Allocate( C, A.nrow, B.ncol ) ;

  // set to 0
  Fill( C, 0 ) ; // 0 int --> convertex 0 double
  
  for ( int k = 0 ; k < A.ncol ; ++k ) // 3
    for ( int i = 0 ; i < A.nrow ; ++i ) // 1
      for ( int j = 0 ; j < B.ncol ; ++j ) // 2
        C.values[i][j] += A.values[i][k]*B.values[k][j] ;
}

void
Mat_x_Mat_321( struct Mat const & A, 
               struct Mat const & B,
               struct Mat       & C ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_321, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  Allocate( C, A.nrow, B.ncol ) ;

  // set to 0
  Fill( C, 0 ) ; // 0 int --> convertex 0 double
  
  for ( int k = 0 ; k < A.ncol ; ++k ) // 3
    for ( int j = 0 ; j < B.ncol ; ++j ) // 2
      for ( int i = 0 ; i < A.nrow ; ++i ) // 1
        C.values[i][j] += A.values[i][k]*B.values[k][j] ;
}

void
Mat_x_Mat_with_tiling( struct Mat const & A, 
                       struct Mat const & B,
                       struct Mat       & C,
                       int tsize ) {
  // check compatibility
  if ( A.ncol != B.nrow ) {
    cerr << "Mat_x_Mat_321, incompatible dimensions\n" ;
    exit(0) ;  
  }
  
  // initialize matrix C
  Allocate( C, A.nrow, B.ncol ) ;

  // set to 0
  Fill( C, 0 ) ; // 0 int --> convertex 0 double
  
  for ( int ii = 0 ; ii < A.nrow ; ii += tsize ) { // 1
    int ii_max = std::min(ii + tsize, A.nrow) ;
    for ( int kk = 0 ; kk < A.ncol ; kk += tsize ) { // 3
      int kk_max = std::min(kk + tsize, A.ncol) ;
      for ( int jj = 0 ; jj < B.ncol ; jj += tsize ) { // 2
        int jj_max = std::min(jj + tsize, B.ncol) ;
        for ( int i = ii ; i < ii_max ; ++i ) // 1
          for ( int k = kk ; k < kk_max ; ++k ) // 3
            for ( int j = jj ; j < jj_max ; ++j ) // 2
              C.values[i][j] += A.values[i][k]*B.values[k][j] ;
      }
    }
  }
}

void
Free( struct Mat & mat ) {
  // check if alreay de-allocated
  if ( mat.memory_block != nullptr ) delete [] mat.memory_block ;
  if ( mat.values       != nullptr ) delete [] mat.values ;
  mat.nrow = 0 ;
  mat.ncol = 0 ;
}
